@@ -26,13 +26,13 @@ parse_response_headers (SoupMessage *req,
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
 	SoupHTTPVersion version;
 
-	g_free((char*)req->reason_phrase);
+	g_free(req->reason_phrase);
 	req->reason_phrase = NULL;
 	if (!soup_headers_parse_response (headers, headers_len,
 					  req->response_headers,
 					  &version,
 					  &req->status_code,
-					  (char **) &req->reason_phrase))
+					  &req->reason_phrase))
 		return SOUP_STATUS_MALFORMED;
 
 	g_object_notify (G_OBJECT (req), SOUP_MESSAGE_STATUS_CODE);
@@ -66,13 +66,19 @@ get_request_headers (SoupMessage *req, GString *header,
 	SoupMessagePrivate *priv = SOUP_MESSAGE_GET_PRIVATE (req);
 	gboolean proxy = GPOINTER_TO_UINT (user_data);
 	SoupURI *uri = soup_message_get_uri (req);
+	char *uri_host;
 	char *uri_string;
 	SoupMessageHeadersIter iter;
 	const char *name, *value;
 
+	if (strchr (uri->host, ':'))
+		uri_host = g_strdup_printf ("[%s]", uri->host);
+	else
+		uri_host = uri->host;
+
 	if (req->method == SOUP_METHOD_CONNECT) {
 		/* CONNECT URI is hostname:port for tunnel destination */
-		uri_string = g_strdup_printf ("%s:%d", uri->host, uri->port);
+		uri_string = g_strdup_printf ("%s:%d", uri_host, uri->port);
 	} else {
 		/* Proxy expects full URI to destination. Otherwise
 		 * just the path.
@@ -86,19 +92,24 @@ get_request_headers (SoupMessage *req, GString *header,
 	} else {
 		g_string_append_printf (header, "%s %s HTTP/1.1\r\n",
 					req->method, uri_string);
-		if (soup_uri_uses_default_port (uri)) {
-			g_string_append_printf (header, "Host: %s\r\n",
-						uri->host);
-		} else {
-			g_string_append_printf (header, "Host: %s:%d\r\n",
-						uri->host, uri->port);
+		if (!soup_message_headers_get (req->request_headers, "Host")) {
+			if (soup_uri_uses_default_port (uri)) {
+				g_string_append_printf (header, "Host: %s\r\n",
+							uri_host);
+			} else {
+				g_string_append_printf (header, "Host: %s:%d\r\n",
+							uri_host, uri->port);
+			}
 		}
 	}
 	g_free (uri_string);
+	if (uri_host != uri->host)
+		g_free (uri_host);
 
 	*encoding = soup_message_headers_get_encoding (req->request_headers);
 	if (*encoding != SOUP_ENCODING_CHUNKED &&
-	    req->request_body->length > 0) {
+	    req->request_body->length > 0 &&
+	    !soup_message_headers_get_content_length (req->request_headers)) {
 		soup_message_headers_set_content_length (req->request_headers,
 							 req->request_body->length);
 	}
