@@ -16,13 +16,13 @@
 #include "soup-auth.h"
 #include "soup-auth-basic.h"
 #include "soup-auth-digest.h"
-#include "soup-auth-manager.h"
 #include "soup-auth-manager-ntlm.h"
 #include "soup-connection.h"
 #include "soup-marshal.h"
 #include "soup-message-private.h"
 #include "soup-message-queue.h"
 #include "soup-session.h"
+#include "soup-session-feature.h"
 #include "soup-session-private.h"
 #include "soup-socket.h"
 #include "soup-ssl.h"
@@ -68,8 +68,6 @@ typedef struct {
 	SoupURI *proxy_uri;
 	SoupAuth *proxy_auth;
 
-	guint max_conns, max_conns_per_host;
-
 	char *ssl_ca_file;
 	SoupSSLCredentials *ssl_creds;
 
@@ -77,12 +75,14 @@ typedef struct {
 
 	char *user_agent;
 
+	GSList *features;
 	SoupAuthManager *auth_manager;
-	SoupAuthManagerNTLM *ntlm_manager;
 
 	GHashTable *hosts; /* SoupURI -> SoupSessionHost */
 	GHashTable *conns; /* SoupConnection -> SoupSessionHost */
 	guint num_conns;
+	guint max_conns, max_conns_per_host;
+	guint io_timeout, idle_timeout;
 
 	/* Must hold the host_lock before potentially creating a
 	 * new SoupSessionHost, or adding/removing a connection.
@@ -91,11 +91,6 @@ typedef struct {
 	GMutex *host_lock;
 
 	GMainContext *async_context;
-
-	/* Holds the timeout value for the connection, when
-	   no response is received.
-	*/
-	guint timeout;
 } SoupSessionPrivate;
 #define SOUP_SESSION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SESSION, SoupSessionPrivate))
 
@@ -107,6 +102,10 @@ static void requeue_message (SoupSession *session, SoupMessage *msg);
 static void cancel_message  (SoupSession *session, SoupMessage *msg,
 			     guint status_code);
 
+static void auth_manager_authenticate (SoupAuthManager *manager,
+				       SoupMessage *msg, SoupAuth *auth,
+				       gboolean retrying, gpointer user_data);
+
 /* temporary until we fix this to index hosts by SoupAddress */
 extern guint     soup_uri_host_hash  (gconstpointer  key);
 extern gboolean  soup_uri_host_equal (gconstpointer  v1,
@@ -121,7 +120,9 @@ extern SoupURI  *soup_uri_copy_root  (SoupURI *uri);
 G_DEFINE_TYPE (SoupSession, soup_session, G_TYPE_OBJECT)
 
 enum {
+	REQUEST_QUEUED,
 	REQUEST_STARTED,
+	REQUEST_UNQUEUED,
 	AUTHENTICATE,
 	LAST_SIGNAL
 };
@@ -139,6 +140,10 @@ enum {
 	PROP_ASYNC_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_USER_AGENT,
+	PROP_IDLE_TIMEOUT,
+	PROP_ADD_FEATURE,
+	PROP_ADD_FEATURE_BY_TYPE,
+	PROP_REMOVE_FEATURE_BY_TYPE,
 
 	LAST_PROP
 };
@@ -163,11 +168,14 @@ soup_session_init (SoupSession *session)
 	priv->max_conns = SOUP_SESSION_MAX_CONNS_DEFAULT;
 	priv->max_conns_per_host = SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT;
 
-	priv->timeout = 0;
-
-	priv->auth_manager = soup_auth_manager_new (session);
+	priv->auth_manager = g_object_new (SOUP_TYPE_AUTH_MANAGER_NTLM,
+					   SOUP_AUTH_MANAGER_NTLM_USE_NTLM, FALSE,
+					   NULL);
+	g_signal_connect (priv->auth_manager, "authenticate",
+			  G_CALLBACK (auth_manager_authenticate), session);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_BASIC);
 	soup_auth_manager_add_type (priv->auth_manager, SOUP_TYPE_AUTH_DIGEST);
+	soup_session_add_feature (session, SOUP_SESSION_FEATURE (priv->auth_manager));
 }
 
 static gboolean
@@ -201,6 +209,9 @@ dispose (GObject *object)
 	soup_session_abort (session);
 	cleanup_hosts (priv);
 
+	while (priv->features)
+		soup_session_remove_feature (session, priv->features->data);
+
 	G_OBJECT_CLASS (soup_session_parent_class)->dispose (object);
 }
 
@@ -218,9 +229,8 @@ finalize (GObject *object)
 
 	g_free (priv->user_agent);
 
-	soup_auth_manager_free (priv->auth_manager);
-	if (priv->ntlm_manager)
-		soup_auth_manager_ntlm_free (priv->ntlm_manager);
+	if (priv->auth_manager)
+		g_object_unref (priv->auth_manager);
 
 	if (priv->proxy_uri)
 		soup_uri_free (priv->proxy_uri);
@@ -254,13 +264,66 @@ soup_session_class_init (SoupSessionClass *session_class)
 
 	/* signals */
 
+	/**
+	 * SoupSession::request-queued:
+	 * @session: the session
+	 * @msg: the request that was queued
+	 *
+	 * Emitted when a request is queued on @session. (Note that
+	 * "queued" doesn't just mean soup_session_queue_message();
+	 * soup_session_send_message() implicitly queues the message
+	 * as well.)
+	 *
+	 * When sending a request, first #SoupSession::request_queued
+	 * is emitted, indicating that the session has become aware of
+	 * the request.
+	 *
+	 * Once a connection is available to send the request on, the
+	 * session emits #SoupSession::request_started. Then, various
+	 * #SoupMessage signals are emitted as the message is
+	 * processed. If the message is requeued, it will emit
+	 * #SoupMessage::restarted, which will then be followed by
+	 * another #SoupSession::request_started and another set of
+	 * #SoupMessage signals when the message is re-sent.
+	 *
+	 * Eventually, the message will emit #SoupMessage::finished.
+	 * Normally, this signals the completion of message
+	 * processing. However, it is possible that the application
+	 * will requeue the message from the "finished" handler (or
+	 * equivalently, from the soup_session_queue_message()
+	 * callback). In that case, the process will loop back to
+	 * #SoupSession::request_started.
+	 *
+	 * Eventually, a message will reach "finished" and not be
+	 * requeued. At that point, the session will emit
+	 * #SoupSession::request_unqueued to indicate that it is done
+	 * with the message.
+	 *
+	 * To sum up: #SoupSession::request_queued and
+	 * #SoupSession::request_unqueued are guaranteed to be emitted
+	 * exactly once, but #SoupSession::request_started and
+	 * #SoupMessage::finished (and all of the other #SoupMessage
+	 * signals) may be invoked multiple times for a given message.
+	 **/
+	signals[REQUEST_QUEUED] =
+		g_signal_new ("request-queued",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0, /* FIXME? */
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      SOUP_TYPE_MESSAGE);
+
 	/**
 	 * SoupSession::request-started:
 	 * @session: the session
 	 * @msg: the request being sent
 	 * @socket: the socket the request is being sent on
 	 *
-	 * Emitted just before a request is sent.
+	 * Emitted just before a request is sent. See
+	 * #SoupSession::request_queued for a detailed description of
+	 * the message lifecycle within a session.
 	 **/
 	signals[REQUEST_STARTED] =
 		g_signal_new ("request-started",
@@ -273,6 +336,26 @@ soup_session_class_init (SoupSessionClass *session_class)
 			      SOUP_TYPE_MESSAGE,
 			      SOUP_TYPE_SOCKET);
 
+	/**
+	 * SoupSession::request-unqueued:
+	 * @session: the session
+	 * @msg: the request that was unqueued
+	 *
+	 * Emitted when a request is removed from @session's queue,
+	 * indicating that @session is done with it. See
+	 * #SoupSession::request_queued for a detailed description of the
+	 * message lifecycle within a session.
+	 **/
+	signals[REQUEST_UNQUEUED] =
+		g_signal_new ("request-unqueued",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_FIRST,
+			      0, /* FIXME? */
+			      NULL, NULL,
+			      soup_marshal_NONE__OBJECT,
+			      G_TYPE_NONE, 1,
+			      SOUP_TYPE_MESSAGE);
+
 	/**
 	 * SoupSession::authenticate:
 	 * @session: the session
@@ -334,6 +417,13 @@ soup_session_class_init (SoupSessionClass *session_class)
 				  G_MAXINT,
 				  SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT,
 				  G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_IDLE_TIMEOUT,
+		g_param_spec_uint (SOUP_SESSION_IDLE_TIMEOUT,
+				   "Idle Timeout",
+				   "Connection lifetime when idle",
+				   0, G_MAXUINT, 0,
+				   G_PARAM_READWRITE));
 	g_object_class_install_property (
 		object_class, PROP_USE_NTLM,
 		g_param_spec_boolean (SOUP_SESSION_USE_NTLM,
@@ -396,6 +486,28 @@ soup_session_class_init (SoupSessionClass *session_class)
 				     "User-Agent string",
 				     NULL,
 				     G_PARAM_READWRITE));
+
+	g_object_class_install_property (
+		object_class, PROP_ADD_FEATURE,
+		g_param_spec_object (SOUP_SESSION_ADD_FEATURE,
+				     "Add Feature",
+				     "Add a feature object to the session",
+				     SOUP_TYPE_SESSION_FEATURE,
+				     G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_ADD_FEATURE_BY_TYPE,
+		g_param_spec_gtype (SOUP_SESSION_ADD_FEATURE_BY_TYPE,
+				    "Add Feature By Type",
+				    "Add a feature object of the given type to the session",
+				    SOUP_TYPE_SESSION_FEATURE,
+				    G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_REMOVE_FEATURE_BY_TYPE,
+		g_param_spec_gtype (SOUP_SESSION_REMOVE_FEATURE_BY_TYPE,
+				    "Remove Feature By Type",
+				    "Remove features of the given type from the session",
+				    SOUP_TYPE_SESSION_FEATURE,
+				    G_PARAM_READWRITE));
 }
 
 static gboolean
@@ -458,15 +570,9 @@ set_property (GObject *object, guint prop_id,
 		priv->max_conns_per_host = g_value_get_int (value);
 		break;
 	case PROP_USE_NTLM:
-		if (g_value_get_boolean (value)) {
-			if (!priv->ntlm_manager)
-				priv->ntlm_manager = soup_auth_manager_ntlm_new (session);
-		} else {
-			if (priv->ntlm_manager) {
-				soup_auth_manager_ntlm_free (priv->ntlm_manager);
-				priv->ntlm_manager = NULL;
-			}
-		}
+		g_object_set_property (G_OBJECT (priv->auth_manager),
+				       SOUP_AUTH_MANAGER_NTLM_USE_NTLM,
+				       value);
 		break;
 	case PROP_SSL_CA_FILE:
 		new_ca_file = g_value_get_string (value);
@@ -493,7 +599,7 @@ set_property (GObject *object, guint prop_id,
 			g_main_context_ref (priv->async_context);
 		break;
 	case PROP_TIMEOUT:
-		priv->timeout = g_value_get_uint (value);
+		priv->io_timeout = g_value_get_uint (value);
 		break;
 	case PROP_USER_AGENT:
 		g_free (priv->user_agent);
@@ -510,7 +616,20 @@ set_property (GObject *object, guint prop_id,
 		} else
 			priv->user_agent = g_strdup (user_agent);
 		break;
+	case PROP_IDLE_TIMEOUT:
+		priv->idle_timeout = g_value_get_uint (value);
+		break;
+	case PROP_ADD_FEATURE:
+		soup_session_add_feature (session, g_value_get_object (value));
+		break;
+	case PROP_ADD_FEATURE_BY_TYPE:
+		soup_session_add_feature_by_type (session, g_value_get_gtype (value));
+		break;
+	case PROP_REMOVE_FEATURE_BY_TYPE:
+		soup_session_remove_feature_by_type (session, g_value_get_gtype (value));
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -533,7 +652,9 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_int (value, priv->max_conns_per_host);
 		break;
 	case PROP_USE_NTLM:
-		g_value_set_boolean (value, priv->ntlm_manager != NULL);
+		g_object_get_property (G_OBJECT (priv->auth_manager),
+				       SOUP_AUTH_MANAGER_NTLM_USE_NTLM,
+				       value);
 		break;
 	case PROP_SSL_CA_FILE:
 		g_value_set_string (value, priv->ssl_ca_file);
@@ -542,12 +663,16 @@ get_property (GObject *object, guint prop_id,
 		g_value_set_pointer (value, priv->async_context ? g_main_context_ref (priv->async_context) : NULL);
 		break;
 	case PROP_TIMEOUT:
-		g_value_set_uint (value, priv->timeout);
+		g_value_set_uint (value, priv->io_timeout);
 		break;
 	case PROP_USER_AGENT:
 		g_value_set_string (value, priv->user_agent);
 		break;
+	case PROP_IDLE_TIMEOUT:
+		g_value_set_uint (value, priv->idle_timeout);
+		break;
 	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
 	}
 }
@@ -629,9 +754,10 @@ free_host (SoupSessionHost *host)
 	g_slice_free (SoupSessionHost, host);
 }	
 
-void
-soup_session_emit_authenticate (SoupSession *session, SoupMessage *msg,
-				SoupAuth *auth, gboolean retrying)
+static void
+auth_manager_authenticate (SoupAuthManager *manager, SoupMessage *msg,
+			   SoupAuth *auth, gboolean retrying,
+			   gpointer session)
 {
 	g_signal_emit (session, signals[AUTHENTICATE], 0, msg, auth, retrying);
 }
@@ -714,59 +840,51 @@ connection_started_request (SoupConnection *conn, SoupMessage *msg,
 					      "User-Agent", priv->user_agent);
 	}
 
+	/* Kludge to deal with the fact that CONNECT msgs come from the
+	 * SoupConnection rather than being queued normally.
+	 */
+	if (msg->method == SOUP_METHOD_CONNECT)
+		g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
+
 	g_signal_emit (session, signals[REQUEST_STARTED], 0,
 		       msg, soup_connection_get_socket (conn));
 }
 
-static void
-find_oldest_connection (gpointer key, gpointer host, gpointer data)
-{
-	SoupConnection *conn = key, **oldest = data;
-
-	/* Don't prune a connection that is currently in use, or
-	 * hasn't been used yet.
-	 */
-	if (soup_connection_is_in_use (conn) ||
-	    soup_connection_last_used (conn) == 0)
-		return;
-
-	if (!*oldest || (soup_connection_last_used (conn) <
-			 soup_connection_last_used (*oldest)))
-		*oldest = conn;
-}
-
-/**
- * soup_session_try_prune_connection:
- * @session: a #SoupSession
- *
- * Finds the least-recently-used idle connection in @session and closes
- * it.
- *
- * Return value: %TRUE if a connection was closed, %FALSE if there are
- * no idle connections.
- **/
 gboolean
 soup_session_try_prune_connection (SoupSession *session)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
-	SoupConnection *oldest = NULL;
+	GPtrArray *conns;
+	GHashTableIter iter;
+	gpointer conn, host;
+	int i;
+
+	conns = g_ptr_array_new ();
 
 	g_mutex_lock (priv->host_lock);
-	g_hash_table_foreach (priv->conns, find_oldest_connection,
-			      &oldest);
-	if (oldest) {
-		/* Ref the connection before unlocking the mutex in
-		 * case someone else tries to prune it too.
+	g_hash_table_iter_init (&iter, priv->conns);
+	while (g_hash_table_iter_next (&iter, &conn, &host)) {
+		/* Don't prune a connection that is currently in use,
+		 * or hasn't been used yet.
 		 */
-		g_object_ref (oldest);
-		g_mutex_unlock (priv->host_lock);
-		soup_connection_disconnect (oldest);
-		g_object_unref (oldest);
-		return TRUE;
-	} else {
-		g_mutex_unlock (priv->host_lock);
+		if (!soup_connection_is_in_use (conn) &&
+		    soup_connection_last_used (conn) > 0)
+			g_ptr_array_add (conns, g_object_ref (conn));
+	}
+	g_mutex_unlock (priv->host_lock);
+
+	if (!conns->len) {
+		g_ptr_array_free (conns, TRUE);
 		return FALSE;
 	}
+
+	for (i = 0; i < conns->len; i++) {
+		soup_connection_disconnect (conns->pdata[i]);
+		g_object_unref (conns->pdata[i]);
+	}
+	g_ptr_array_free (conns, TRUE);
+
+	return TRUE;
 }
 
 static void
@@ -934,7 +1052,8 @@ soup_session_get_connection (SoupSession *session, SoupMessage *msg,
 		SOUP_CONNECTION_PROXY_URI, priv->proxy_uri,
 		SOUP_CONNECTION_SSL_CREDENTIALS, priv->ssl_creds,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
-		SOUP_CONNECTION_TIMEOUT, priv->timeout,
+		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
+		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
 		NULL);
 	g_signal_connect (conn, "connect_result",
 			  G_CALLBACK (connect_result),
@@ -982,6 +1101,8 @@ message_finished (SoupMessage *msg, gpointer user_data)
 	if (!SOUP_MESSAGE_IS_STARTING (msg)) {
 		soup_message_queue_remove_message (priv->queue, msg);
 		g_signal_handlers_disconnect_by_func (msg, message_finished, session);
+		g_signal_handlers_disconnect_by_func (msg, redirect_handler, session);
+		g_signal_emit (session, signals[REQUEST_UNQUEUED], 0, msg);
 	}
 }
 
@@ -1002,6 +1123,8 @@ queue_message (SoupSession *session, SoupMessage *msg,
 
 	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
 	soup_message_queue_append (priv->queue, msg);
+
+	g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
 }
 
 /**
@@ -1210,3 +1333,62 @@ soup_session_abort (SoupSession *session)
 
 	g_slist_free (conns);
 }
+
+void
+soup_session_add_feature (SoupSession *session, SoupSessionFeature *feature)
+{
+	SoupSessionPrivate *priv;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (SOUP_IS_SESSION_FEATURE (feature));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+	priv->features = g_slist_prepend (priv->features, g_object_ref (feature));
+	soup_session_feature_attach (feature, session);
+}
+
+void
+soup_session_add_feature_by_type (SoupSession *session, GType feature_type)
+{
+	SoupSessionFeature *feature;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+	g_return_if_fail (g_type_is_a (feature_type, SOUP_TYPE_SESSION_FEATURE));
+
+	feature = g_object_new (feature_type, NULL);
+	soup_session_add_feature (session, feature);
+	g_object_unref (feature);
+}
+
+void
+soup_session_remove_feature (SoupSession *session, SoupSessionFeature *feature)
+{
+	SoupSessionPrivate *priv;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+	if (g_slist_find (priv->features, feature)) {
+		priv->features = g_slist_remove (priv->features, feature);
+		soup_session_feature_detach (feature, session);
+		g_object_unref (feature);
+	}
+}
+
+void
+soup_session_remove_feature_by_type (SoupSession *session, GType feature_type)
+{
+	SoupSessionPrivate *priv;
+	GSList *f;
+
+	g_return_if_fail (SOUP_IS_SESSION (session));
+
+	priv = SOUP_SESSION_GET_PRIVATE (session);
+restart:
+	for (f = priv->features; f; f = f->next) {
+		if (G_TYPE_CHECK_INSTANCE_TYPE (f->data, feature_type)) {
+			soup_session_remove_feature (session, f->data);
+			goto restart;
+		}
+	}
+}
